package de.elster.lohn.util.eTIN;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.DataFormatException;

public class eTin {

	/**
	 * die Variablen mit *Mod4 enthalten die jeweils schon in die
	 * 4er-Zeichenkette umgewandelten Strings von Nachname und Vorname das
	 * Geburtsdatum ist bereits umgewandelt in Jahr - Monatsbuchstabe - Tag
	 * (z.B. "00A01" fr "01.01.2000")
	 */

	private String nachnameMod4;

	private String vornameMod4;

	private String geburtsDatum;

	/**
	 * Pruefziffer der eTIN
	 */
	private String pruefZiffer;

	/**
	 * Dieser String enthaelt alle erlaubten "normalen" Buchstaben sowohl in
	 * Klein- als auch Grossschreibung
	 */
	private static String erlaubteBuchstabenStr;

	/**
	 * Diese Tabelle enthaelt die zu ersetzenden Zeichen
	 */
	private static Hashtable<String,String> ersatzListe;

	/**
	 * Hier werden nur die Vokale gespeichert
	 */
	private static ArrayList<String> vokalListe;

	/**
	 * Die Umrechnungstabelle, um die Monatszahl in einen Buchstaben zu wandeln
	 */
	private static String monatsBuchstabenListe;

	/**
	 * Tabelle mit den Pruefzahlen fuer alle geraden TIN-Zeichen
	 */
	private static Hashtable<String,Integer> pruefzifferListeGerade;

	/**
	 * Tabelle mit den Pruefzahlen fuer alle ungeraden TIN-Zeichen
	 */
	private static Hashtable<String,Integer> pruefzifferListeUngerade;

	/**
	 * Instanziiert die Klasse eTIN und ruft mit den bergebenen Parametern die
	 * Methode setDaten(...) auf.
	 * 
	 * @param newNachname
	 *            der Familienname
	 * @param newVorname
	 *            der Vorname
	 * @param newGebDatum
	 *            das Geburtsdatum (YYYYMMTT)
	 */
	public eTin(String newNachname, String newVorname, Integer newGebDatum) {
		setDaten(newNachname, newVorname, newGebDatum);
	}

	static {
		erlaubteBuchstabenStr = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";

		// Ersatzliste lt. Anforderung
		ersatzListe = new Hashtable<String,String>(109);
		ersatzListe.put("", "A"); // 192
		ersatzListe.put("", "A"); // 193
		ersatzListe.put("", "A"); // 194
		ersatzListe.put("", "A"); // 195
		ersatzListe.put("", "A"); // 197
		ersatzListe.put("", "A"); // 224
		ersatzListe.put("", "A"); // 225
		ersatzListe.put("", "A"); // 226
		ersatzListe.put("", "A"); // 227
		ersatzListe.put("", "A"); // 229
		ersatzListe.put("", "AE"); // 196
		ersatzListe.put("", "AE"); // 228
		ersatzListe.put("", "AE"); // 198
		ersatzListe.put("", "AE"); // 230
		ersatzListe.put("", "C"); // 231
		ersatzListe.put("", "C"); // 199
		ersatzListe.put("", "D"); // 208
		ersatzListe.put("", "D"); // 240
		ersatzListe.put("", "E"); // 200
		ersatzListe.put("", "E"); // 201
		ersatzListe.put("", "E"); // 202
		ersatzListe.put("", "E"); // 203
		ersatzListe.put("", "E"); // 232
		ersatzListe.put("", "E"); // 233
		ersatzListe.put("", "E"); // 234
		ersatzListe.put("", "E"); // 235
		ersatzListe.put("", "I"); // 204
		ersatzListe.put("", "I"); // 205
		ersatzListe.put("", "I"); // 206
		ersatzListe.put("", "I"); // 207
		ersatzListe.put("", "I"); // 236
		ersatzListe.put("", "I"); // 237
		ersatzListe.put("", "I"); // 238
		ersatzListe.put("", "I"); // 239
		ersatzListe.put("", "N"); // 209
		ersatzListe.put("", "N"); // 241
		ersatzListe.put("", "O"); // 210
		ersatzListe.put("", "O"); // 211
		ersatzListe.put("", "O"); // 212
		ersatzListe.put("", "O"); // 213
		ersatzListe.put("", "OE"); // 214
		ersatzListe.put("", "O"); // 216
		ersatzListe.put("", "O"); // 242
		ersatzListe.put("", "O"); // 243
		ersatzListe.put("", "O"); // 244
		ersatzListe.put("", "O"); // 245
		ersatzListe.put("", "OE"); // 189
		ersatzListe.put("", "O"); // 248
		ersatzListe.put("", "S"); // 166
		ersatzListe.put("", "S"); // 168
		ersatzListe.put("", "U"); // 217
		ersatzListe.put("", "U"); // 218
		ersatzListe.put("", "U"); // 219
		ersatzListe.put("", "U"); // 249
		ersatzListe.put("", "U"); // 250
		ersatzListe.put("", "U"); // 251
		ersatzListe.put("", "UE"); // 220
		ersatzListe.put("", "UE"); // 252
		ersatzListe.put("", "Y"); // 190
		ersatzListe.put("", "Y"); // 221
		ersatzListe.put("", "Y"); // 253
		ersatzListe.put("", "Y"); // 255
		ersatzListe.put("", "Z"); // 180
		ersatzListe.put("", "Z"); // 184

		ersatzListe.put("", "SS"); // 223
		// die klein geschriebenen Buchstaben werden allerdings auch durch
		// gross geschriebene Buchstaben ersetzt
		for (int i = 0; i < erlaubteBuchstabenStr.length(); i++)
			ersatzListe.put(erlaubteBuchstabenStr.substring(i, i + 1),
					erlaubteBuchstabenStr.substring(i, i + 1).toUpperCase());

		vokalListe = new ArrayList<String>(5);
		vokalListe.add("A");
		vokalListe.add("E");
		vokalListe.add("I");
		vokalListe.add("O");
		vokalListe.add("U");

		// Liste der Buchstaben, die einem Monat entsprechen, z.B. Januar = 0 =
		// A;
		monatsBuchstabenListe = "ABCDEFGHIJKL";

		// Tabelle mit Pruefziffernwerten lt. Anforderung
		pruefzifferListeGerade = new Hashtable<String,Integer>(36);
		pruefzifferListeUngerade = new Hashtable<String,Integer>(36);
		// Pruefziffernwerte fuer "gerade" eTIN-Position
		pruefzifferListeGerade.put("0", new Integer(0));
		pruefzifferListeGerade.put("1", new Integer(1));
		pruefzifferListeGerade.put("2", new Integer(2));
		pruefzifferListeGerade.put("3", new Integer(3));
		pruefzifferListeGerade.put("4", new Integer(4));
		pruefzifferListeGerade.put("5", new Integer(5));
		pruefzifferListeGerade.put("6", new Integer(6));
		pruefzifferListeGerade.put("7", new Integer(7));
		pruefzifferListeGerade.put("8", new Integer(8));
		pruefzifferListeGerade.put("9", new Integer(9));
		pruefzifferListeGerade.put("A", new Integer(0));
		pruefzifferListeGerade.put("B", new Integer(1));
		pruefzifferListeGerade.put("C", new Integer(2));
		pruefzifferListeGerade.put("D", new Integer(3));
		pruefzifferListeGerade.put("E", new Integer(4));
		pruefzifferListeGerade.put("F", new Integer(5));
		pruefzifferListeGerade.put("G", new Integer(6));
		pruefzifferListeGerade.put("H", new Integer(7));
		pruefzifferListeGerade.put("I", new Integer(8));
		pruefzifferListeGerade.put("J", new Integer(9));
		pruefzifferListeGerade.put("K", new Integer(10));
		pruefzifferListeGerade.put("L", new Integer(11));
		pruefzifferListeGerade.put("M", new Integer(12));
		pruefzifferListeGerade.put("N", new Integer(13));
		pruefzifferListeGerade.put("O", new Integer(14));
		pruefzifferListeGerade.put("P", new Integer(15));
		pruefzifferListeGerade.put("Q", new Integer(16));
		pruefzifferListeGerade.put("R", new Integer(17));
		pruefzifferListeGerade.put("S", new Integer(18));
		pruefzifferListeGerade.put("T", new Integer(19));
		pruefzifferListeGerade.put("U", new Integer(20));
		pruefzifferListeGerade.put("V", new Integer(21));
		pruefzifferListeGerade.put("W", new Integer(22));
		pruefzifferListeGerade.put("X", new Integer(23));
		pruefzifferListeGerade.put("Y", new Integer(24));
		pruefzifferListeGerade.put("Z", new Integer(25));
		// Pruefziffernwerte fuer "ungerade" eTIN-Position
		pruefzifferListeUngerade.put("0", new Integer(1));
		pruefzifferListeUngerade.put("1", new Integer(0));
		pruefzifferListeUngerade.put("2", new Integer(5));
		pruefzifferListeUngerade.put("3", new Integer(7));
		pruefzifferListeUngerade.put("4", new Integer(9));
		pruefzifferListeUngerade.put("5", new Integer(13));
		pruefzifferListeUngerade.put("6", new Integer(15));
		pruefzifferListeUngerade.put("7", new Integer(17));
		pruefzifferListeUngerade.put("8", new Integer(19));
		pruefzifferListeUngerade.put("9", new Integer(21));
		pruefzifferListeUngerade.put("A", new Integer(1));
		pruefzifferListeUngerade.put("B", new Integer(0));
		pruefzifferListeUngerade.put("C", new Integer(5));
		pruefzifferListeUngerade.put("D", new Integer(7));
		pruefzifferListeUngerade.put("E", new Integer(9));
		pruefzifferListeUngerade.put("F", new Integer(13));
		pruefzifferListeUngerade.put("G", new Integer(15));
		pruefzifferListeUngerade.put("H", new Integer(17));
		pruefzifferListeUngerade.put("I", new Integer(19));
		pruefzifferListeUngerade.put("J", new Integer(21));
		pruefzifferListeUngerade.put("K", new Integer(2));
		pruefzifferListeUngerade.put("L", new Integer(4));
		pruefzifferListeUngerade.put("M", new Integer(18));
		pruefzifferListeUngerade.put("N", new Integer(20));
		pruefzifferListeUngerade.put("O", new Integer(11));
		pruefzifferListeUngerade.put("P", new Integer(3));
		pruefzifferListeUngerade.put("Q", new Integer(6));
		pruefzifferListeUngerade.put("R", new Integer(8));
		pruefzifferListeUngerade.put("S", new Integer(12));
		pruefzifferListeUngerade.put("T", new Integer(14));
		pruefzifferListeUngerade.put("U", new Integer(16));
		pruefzifferListeUngerade.put("V", new Integer(10));
		pruefzifferListeUngerade.put("W", new Integer(22));
		pruefzifferListeUngerade.put("X", new Integer(23));
		pruefzifferListeUngerade.put("Y", new Integer(24));
		pruefzifferListeUngerade.put("Z", new Integer(25));
	}

	/**
	 * Setzt den Namen fr die eTIN-Berechnung.
	 * 
	 * @param aNachname
	 *            der Nachname
	 */
	public void setGebName(String aNachname) {
		if (aNachname == null)
			aNachname = "";
		nachnameMod4 = getViererZeichen(zeichenErsetzen(aNachname));

	}

	/**
	 * Setzt das Geburtsdatum fr die eTIN-Berechnung. Ein ggfls. ungltiges
	 * Datum wird durch 01.01.2000 ersetzt.
	 * 
	 * @param aGebDatum
	 *            das Geburtdatum im Format dd.mm.yy bzw. dd.mm.yyyy D
	 */
	public void setGebDatum(Integer aGebDatum) {
		int gebDatum;
		if (aGebDatum == null)
			gebDatum = new Integer("20000101").intValue();
		else {
			gebDatum = aGebDatum.intValue();
			korrigiereDatum(gebDatum);
		}

	}

	/**
	 * Setzt den Vornamen fr die eTIN-Berechnung.
	 * 
	 * @param aVorname
	 *            Vorname
	 */
	public void setVorname(String aVorname) {
		if (aVorname == null)
			aVorname = "";
		vornameMod4 = getViererZeichen(zeichenErsetzen(aVorname));

	}

	/**
	 * Setzt alle notwendigen Daten fr die eTIN-Berechnung.
	 * 
	 * @param newNachname
	 *            den Familiennamen
	 * @param newVorname
	 *            der Vorname
	 * @param newGebDatum
	 *            das Geburtsdatum im Format dd.mm.yy bzw dd.mm.yyyy
	 */
	public void setDaten(String newNachname, String newVorname,
			Integer newGebDatum) {
		setGebName(newNachname);
		setVorname(newVorname);
		setGebDatum(newGebDatum);

	}

	/**
	 * lscht alle Daten
	 */
	public void clearDaten() {
		setDaten("", "", new Integer(0));
	}

	/**
	 * 
	 * @return String 14stellige-eTIN
	 */
	public String toString() {
		return geteTIN();
		// return getGebNameNormalisiert() + ";" + getVornameNormalisiert() +
		// ";" + getGebDatum();
	}

	/**
	 * @param originalString
	 * @return String
	 */
	private String zeichenErsetzen(String originalString) {

		// Erste Umwandlung: Leerzeichen vorne und hinten killen, danach alle
		// Zeichen in Grossbuchstaben wandeln, soweit moeglich.
		originalString = originalString.trim().toUpperCase();

		StringBuffer sb = new StringBuffer();
		String ergeb;

		for (int i = 0; i < originalString.length(); i++) {
			ergeb = (String) ersatzListe.get(String.valueOf(originalString
					.charAt(i)));
			if (ergeb != null) {
				sb.append(ergeb);
			}
		}
		return sb.toString();
	}

	/**
	 * Erzeugt aus einem String wie z.B. Nachname die lt. Anforderung
	 * benoetigten 4 Zeichen zum Aufbau der eTIN
	 * 
	 * @param str
	 * @return String
	 */
	private String getViererZeichen(String str) {
		StringBuffer erg = new StringBuffer();

		// Zuerst einmal alle Sonderzeichen umwandeln bzw. ersetzen
		StringBuffer string = new StringBuffer(str);

		// Dann jedes SCH durch 'Y' ersetzen, lt. Anforderung
		int p;
		while ((p = string.indexOf("SCH")) >= 0) {
			string.replace(p, p + 3, "Y");
		}

		// Die Konsonanten von links nach rechts aufbauen,
		// wobei maximal 4 benoetigt werden
		for (int i = 0; i < string.length() && erg.length() < 4; i++) {
			if (isKonsonant(string.charAt(i))) {
				erg.append(string.charAt(i));
			}
		}

		if (erg.length() < 4) {
			// Wenn nicht mehr genuegend Konsonanten zur Verfuegung standen,
			// dann den Rest mit Vokalen auffuellen, die von der Zeichenkette
			// von rechts nach links ermittelt werden
			for (int i = string.length() - 1; i >= 0 && erg.length() < 4; i--) {
				if (isVokal(string.charAt(i))) {
					erg.append(string.charAt(i));
				}
			}

			// Wenn auch dann immer noch keine 4 Zeichen zusammengekommen sind,
			// dann den Rest mit 'X' auffuellen
			while (erg.length() < 4) {
				erg.append("X");
			}
		}

		return erg.toString();
	}

	/**
	 * @param c
	 * @return boolean
	 */
	private boolean isVokal(char c) {
		// Der Leerstring sorgt dafuer, dass auch dem (char)c ein (String)c
		// wird.
		return vokalListe.contains("" + c);
	}

	/**
	 * Wenn es kein Vokal ist, dann muss es ein Konsonant sein
	 * 
	 * @param c
	 * @return boolean
	 */
	private boolean isKonsonant(char c) {
		return !isVokal(c);
	}

	/**
	 * Setzt ungltige Geburtsdaten auf den Default-Wert "00A01" (01.01.2000)
	 */
	private void setDefaultGebDatum() {
		geburtsDatum = "00A01";

	}

	/**
	 * Einfache Datumspruefung per Abfangen einer Exception
	 */
	private void korrigiereDatum(int gebDat) {

		int gebDatumJahr2 = gebDat / 10000 - (((gebDat / 10000) / 100) * 100);
		int gebDatumMonat = (gebDat - ((gebDat / 10000) * 10000)) / 100;
		int gebDatumTag = gebDat
				- (((gebDat / 10000) * 10000) + (gebDatumMonat * 100));

		if ((gebDatumTag <= 0) || (gebDatumTag >= 32)) {
			setDefaultGebDatum();

		} else

		if ((gebDatumMonat <= 0) || (gebDatumMonat >= 13)) {
			setDefaultGebDatum();
		} else {

			geburtsDatum = intToString(gebDatumTag) + "."
					+ intToString(gebDatumMonat) + "."
					+ intToString(gebDatumJahr2);
			DateFormat df = new SimpleDateFormat("dd.MM.yy");

			try {
				df.setLenient(false);
				df.parse(geburtsDatum);
				geburtsDatum = "" + intToString(gebDatumJahr2)
						+ monatsBuchstabenListe.charAt(gebDatumMonat - 1)
						+ intToString(gebDatumTag);
			} catch (ParseException e) {
				setDefaultGebDatum();
			}

		}
	}

	/**
	 * Wandelt 1- oder 2-stellige integer-Werte in 2-stellige Strings
	 * 
	 * @param wert
	 * @return String
	 */

	private String intToString(int wert) {
		StringBuffer sb = new StringBuffer(String.valueOf(wert));
		while (sb.length() < 2) {
			sb.insert(0, "0");
		}
		return sb.toString();

	}

	/**
	 * @roseuid 3E6DC4510122
	 */
	private static String berechnePruefziffer(String s) {
		// String s = nachnameMod4 + vornameMod4 + gebDatum;
		String zeichen;
		int wert = 0;
		int restwert;

		// Je nach gerader oder ungerader Position des Zeichens in
		// der eTIN wird in der entsprechenden Tabelle nachgeschlagen.
		for (int i = 1; i <= s.length(); i++) {
			zeichen = "" + s.charAt(i - 1);
			if (i % 2 == 0) {
				wert += ((Integer) pruefzifferListeGerade.get(zeichen))
						.intValue();
			} else {
				wert += ((Integer) pruefzifferListeUngerade.get(zeichen))
						.intValue();
			}
		}
		// Alle Pruefziffern zusammengezaehlt Modulo 26 ergibt einen
		// Restwert, der direkt in ASCII umgewandelt wird, d.h. 65 = 'A'
		restwert = wert % 26;
		// this.pruefZiffer = "" + new Character((char) (restwert + 65));
		return "" + new Character((char) (restwert + 65));
	}

	/**
	 * Diese gibt die berechnete eTIN zurueck (14 Stellen)
	 * 
	 * @return String 14stellige eTIN
	 */
	public String geteTIN() {
		this.pruefZiffer = berechnePruefziffer(this.nachnameMod4
				+ this.vornameMod4 + this.geburtsDatum);
		return this.nachnameMod4 + this.vornameMod4 + this.geburtsDatum
				+ this.pruefZiffer;

	}

	/**
	 * Statische Methode mit der direkt die eTIN berechnet und zurckgegeben
	 * wird, ohne eine Klasse eTin zu instanzieren.
	 * 
	 * @param newNachname
	 *            der Familienname
	 * @param newVorname
	 *            der Vorname
	 * @param newGebDatum
	 *            im Format YYYYMMDD
	 * @return String 14stellige eTIN
	 */
	public static String berechneETIN(String newNachname, String newVorname,
			Integer newGebDatum) {
		eTin e = new eTin(newNachname, newVorname, newGebDatum);
		return e.geteTIN();
	}

	/**
	 * Statische Methode mit der direkt die eTIN berechnet und zurckgegeben
	 * wird, ohne eine Klasse eTin zu instanzieren.
	 * 
	 * @param newNachname
	 *            der Familienname
	 * @param newVorname
	 *            der Vorname
	 * @param newGebDatum
	 *            im Format dd.mm.yy bzw. ddmmyyyy
	 * @return String 14stellige eTIN
	 */
	public static String berechneETIN(String newGebName, String newVorname,
			String newGebDatum) {
		Integer newDatum;
		switch (newGebDatum.trim().length()) {
		case 8: {
			newDatum = new Integer("19" + newGebDatum.substring(6)
					+ newGebDatum.substring(3, 5) + newGebDatum.substring(0, 2));

		}
			break;
		case 10: {
			newDatum = new Integer(newGebDatum.substring(6)
					+ newGebDatum.substring(3, 5) + newGebDatum.substring(0, 2));
		}
			break;
		default: {
			newDatum = new Integer("20000101");
		}

		}

		eTin e = new eTin(newGebName, newVorname, newDatum);
		return e.geteTIN();
	}
	/**
	 * Statische Methode mit der direkt die eTIN berechnet und zurckgegeben
	 * wird, ohne eine Klasse eTin zu instanzieren.
	 * 
	 * @param newNachname
	 *            der Familienname
	 * @param newVorname
	 *            der Vorname
	 * @param newGebDatum
	 *            YYYYMMDD
	 * @return String 14stellige eTIN
	 */
	//Methode wird bentigt fr etin-berechnung in DRVSimpleTypeHandler
	public static String berechneETINyyyymmdd(String newGebName, String newVorname,
			String newGebDatum) {
		Integer newDatum;
		switch (newGebDatum.trim().length()) {
		case 8: 
		{
			newDatum = new Integer(newGebDatum);
			break;

		}
		default: {
			newDatum = new Integer("20000101");
		}

		}

		eTin e = new eTin(newGebName, newVorname, newDatum);
		return e.geteTIN();
	}

	/**
	 * Statische Methode mit der das Format einer bergebenen eTin auf
	 * Gltigkeit geprft wird. Geprft wird gegen das Pattern sowie gegen die
	 * Prfziffer. Entspricht die eTin nicht dem Pattern, so fliegt eine
	 * DataFormatException.
	 * 
	 * @param etin
	 * @return boolean (returns false wenn die bergebene eTin kein gltiges
	 *         Format hat)
	 */

	public static boolean pruefeFormat(String etin) throws DataFormatException {
		Pattern p;
		Matcher m;
		p = Pattern.compile("[A-Z]{8}\\d{2}[A-L]{1}[0-3][0-9][A-Z]{1}");
		m = p.matcher(etin);
		if (m.matches() == false) {
			throw new DataFormatException("Die eTIN =" + etin
					+ " entspricht nicht dem Pattern " + p.pattern());
		} else
			return berechnePruefziffer(etin.substring(0, 13)).equals(
					etin.substring(13));
	}

}
